"""

"""
import numpy as np
import pandas as pd
import networkx as nx
import multiprocessing
from multiprocessing import Queue, JoinableQueue, Process
import inspect
from pandas.api.types import is_numeric_dtype
from navipy.database import DataBaseLoad
import navipy.moving.maths as navimomath


def defaultcallback(database, posorients):
    raise NameError('No Callback')


class AbstractAgent():
    """
    An abtract class for agent
    """

    def __init__(self,
                 database_filename,
                 memory_friendly=False):

        self.db = DataBaseLoad(database_filename)
        self.dbname = database_filename
        if memory_friendly:
            self.__posorients = None
        else:
            self.__posorients = self.db.get_posorients()
        # set mode of motion
        mode_move = {'mode': 'on_cubic_grid',
                     'param': {'grid_spacing':
                               pd.Series(data=1,
                                         index=['dx', 'dy', 'dz'])}}
        self.mode_of_motion = mode_move

    @property
    def posorients(self):
        toreturn = self.__posorients
        if toreturn is not None:
            toreturn = toreturn.copy()
        return toreturn

    @property
    def mode_of_motion(self):
        """
        """
        toreturn = self.__mode_move
        toreturn['describe'] = \
            navimomath.mode_moves_supported()[
                self.__mode_move['mode']]['describe']
        return toreturn

    @mode_of_motion.setter
    def mode_of_motion(self, mode):
        """

        """
        if not isinstance(mode, dict):
            raise TypeError('Mode is not a dictionary')
        if 'mode' not in mode:
            raise KeyError("'mode' is not a key of mode")
        if 'param' not in mode:
            raise KeyError("'param' is not a key of mode")
        if mode['mode'] in navimomath.mode_moves_supported().keys():
            for param in navimomath.mode_moves_supported()[
                    mode['mode']]['param']:
                if param not in mode['param']:
                    raise KeyError(
                        "'{}' is not in mode['param']".format(param))
            self.__mode_move = mode
        else:
            raise ValueError('mode is not supported')

    def abstractmove(self, posorients_vel):
        if isinstance(posorients_vel, pd.Series) is False:
            raise TypeError('posorients_vel should be a pandas Series')
        for col in ['x', 'y', 'z', 'alpha_0', 'alpha_1', 'alpha_2',
                    'dx', 'dy', 'dz', 'dalpha_0', 'dalpha_1', 'dalpha_2']:
            if col not in posorients_vel.index:
                raise KeyError(
                    'posorients_vel should have {} as index'.format(col))
        # Compute the next position
        posorients_vel = navimomath.next_pos(
            posorients_vel,
            move_mode=self.__mode_move['mode'],
            move_param=self.__mode_move['param'])

        # Compute the closest possible position
        if posorients_vel is None:
            tmp = navimomath.closest_pos_memory_friendly(
                posorients_vel,
                self.db)
            posorients_vel[['x', 'y', 'z',
                            'alpha_0', 'alpha_1', 'alpha_2']] = tmp
            posorients_vel.name = tmp.name
        else:
            tmp = navimomath.closest_pos(
                posorients_vel,
                self.__posorients)
            posorients_vel[['x', 'y', 'z',
                            'alpha_0', 'alpha_1', 'alpha_2']] = tmp
            posorients_vel.name = tmp.name
        return posorients_vel


class Single(AbstractAgent, Process):

    def __init__(self,
                 database_filename,
                 initial_condition,
                 memory_friendly=False,
                 posorients_queue=None,
                 results_queue=None):
        if (posorients_queue is not None) and (results_queue is not None):
            multiprocessing.Process.__init__(self)
        AbstractAgent.__init__(self, database_filename,
                               memory_friendly)

        self.__posorientvel = pd.Series(
            data=0,
            index=['x', 'y', 'z',
                   'alpha_0', 'alpha_1', 'alpha_2',
                   'dx', 'dy', 'dz',
                   'dalpha_0', 'dalpha_1', 'dalpha_2'],
            dtype=np.float)

        if isinstance(initial_condition, pd.Series):
            if is_numeric_dtype(initial_condition):
                common_id = list(set(initial_condition.index).intersection(
                    self.__posorientvel.index))
                self.__posorientvel.loc[common_id] = \
                    initial_condition.loc[common_id]
            else:
                raise TypeError('vel should be numeric')

        else:
            raise TypeError('vel should be a pandas Series')

        self.__posorients_queue = posorients_queue
        self.__results_queue = results_queue
        self.__callback_function = defaultcallback

    def move(self):
        # Compute the next position
        tmp = self.__callback_function(database=self.db,
                                       posorient=self.__posorientvel)
        common_id = list(set(tmp.index).intersection(
            self.__posorientvel.index))
        self.__posorientvel.loc[common_id] = tmp.loc[common_id]
        self.__posorientvel = self.abstractmove(self.__posorientvel)

    def fly(self,  nsteps):
        """move until either speed is null, or nsteps has been reached"""
        prev_move = self.__posorientvel
        for stepi in range(nsteps):
            self.move()
            if prev_move.equals(self.__posorientvel):
                break
            prev_move = self.__posorientvel

    def run(self):
        """ Only supported when multiprocess"""
        if self.__posorients_queue is None or self.__results_queue is None:
            raise NameError('Single agent class has not be inititialised '
                            + 'with multiprocessing suppport')
        proc_name = self.name
        print('Process {} started'.format(proc_name))
        while True:
            start_posorient = self.__posorients_queue.get(timeout=1)
            if start_posorient is None:
                # Poison pill means shutdown)
                break
            common_id = list(set(start_posorient.index).intersection(
                self.__posorientvel.index))
            self.__posorientvel.loc[common_id] = start_posorient.loc[common_id]
            self.move()
            next_posorient = self.__posorientvel

            self.__posorients_queue.task_done()
            self.__results_queue.put((start_posorient, next_posorient))
        self.__posorients_queue.task_done()
        print('Process {} done'.format(proc_name))

    @property
    def callback_function(self):
        return inspect.getsourcelines(self.__callback_function)

    @callback_function.setter
    def callback_function(self, callback_function):
        self.__callback_function = callback_function

    @property
    def position(self):
        return self.__posorientvel.loc[['x', 'y', 'z']]

    @property
    def velocity(self):
        return self.__posorientvel.loc[['dx', 'dy', 'dz']]

    @property
    def orientation(self):
        return self.__posorientvel.loc[['alpha_0', 'alpha_1', 'alpha_2']]

    @property
    def angular_velocity(self):
        return self.__posorientvel.loc[['dalpha_0', 'dalpha_1', 'dalpha_2']]


class Multi(AbstractAgent):

    def __init__(self, database_filename):
        super().__init__(database_filename, False)
        # Init the graph
        self.__graph = nx.DiGraph()
        for row_id, posor in self.db.get_posorients().iterrows():
            posor.name = row_id
            self.__graph.add_node(row_id,
                                  posorient=posor)

    @property
    def graph(self):
        return self.__graph

    @graph.setter
    def graph(self, graph):
        if isinstance(graph, nx.DiGraph) is False:
            raise TypeError('graph is not a nx.DiGraph')
        self.__graph = graph.copy()
        self.check_graph()

    def build_graph(self, callback_function,
                    ncpu=5,
                    timeout=1):
        # Build a list of nodes
        results_edges = []
        posorients_queue = JoinableQueue()
        results_queue = Queue()
        for node in self.__graph.nodes:
            posorients_queue.put(self.__graph.nodes[node]['posorient'])
            initpos = 0 * self.__graph.nodes[node]['posorient']

        # Start ndatabase loader
        num_agents = ncpu
        agents = [Single(self.dbname,
                         initial_condition=initpos,
                         memory_friendly=False,
                         posorients_queue=posorients_queue,
                         results_queue=results_queue)
                  for _ in range(num_agents)]
        for w in agents:
            w.callback_function = callback_function
            w.start()

        # Add a poison pill for each agent
        for _ in range(num_agents):
            posorients_queue.put(None)

        # Wait for all of the tasks to finish
        # posorients_queue.join()

        for _ in range(nx.number_of_nodes(self.__graph)):
            result = results_queue.get(timeout=timeout)
            results_edges.append((result[0].name,
                                  result[1].name))
            # print(results_edges[-1])
        self.__graph.add_edges_from(results_edges)
        self.check_graph()

    def check_graph(self):
        self.check_single_target()

    def check_single_target(self):
        for node in self.__graph.nodes:
            # not connected -> loop not ran
            for count, _ in enumerate(self.__graph.neighbors(node)):
                # count == 0 -> connected to one node
                # count == 1 -> connected to two nodes
                if count > 0:
                    raise ValueError(
                        'Node {} leads to several locations'.format(node))

    def find_attractors(self):
        """Return a list of node going to each attractor in a graph
        """
        attractors = list()
        for attractor in nx.attracting_components(self.__graph):
            att = dict()
            att['attractor'] = attractor
            attractors.append(att)
        return attractors

    def find_attractors_sources(self, attractors=None):
        """Find all sources going to each attractors
        """
        if attractors is None:
            attractors = self.find_attractors()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'attractor' not in keyatt:
                raise ValueError(
                    'Each attractors should contain the key attractor')

        # Calculate connection
        for att_i, att in enumerate(attractors):

            # [0] because one node of the attractor is enough
            # all other node of the attractor are connected to this one
            target = list(att['attractor'])[0]
            attractors[att_i]['paths'] = \
                nx.shortest_path(self.graph, target=target)
            attractors[att_i]['sources'] = \
                list(attractors[att_i]['paths'].keys())
        return attractors

    def catchment_area(self, attractors=None):
        """Return the catchment area for attractors
        """
        if attractors is None:
            attractors = self.find_attractors_sources()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'sources' not in keyatt:
                raise ValueError(
                    'Each attractors should contains a list of sources')

        return [len(att['sources']) for att in attractors]

    def reach_goals(self, goals):
        """ Return all paths to the goals """
        return nx.shortest_path(self.__graph, target=goals)

    def neighboring_nodes(self, target):
        """ Return the nodes going to the target """
        # Reverse graph because nx.neighbors give the end node
        # and we want to find the start node going to target
        # not where target goes.
        tmpgraph = self.__graph.reverse(copy=True)
        neighbors = tmpgraph.neighbors(target)
        return neighbors
