import unittest
import sqlite3
import numpy as np
import pandas as pd
import navipy.database as database
import navipy.processing.pcode as pcode
from navipy.processing.tools import is_numeric_array
import pkg_resources


class TestCase(unittest.TestCase):
    def setUp(self):
        self.mydb_filename = pkg_resources.resource_filename(
            'navipy', 'resources/database.db')
        self.mydb = database.DataBaseLoad(self.mydb_filename)

    def test_scene_posorient(self):
        conn = sqlite3.connect(self.mydb_filename)
        c = conn.cursor()
        c.execute(""" SELECT * FROM position_orientation WHERE (rowid=1) """)
        rows = c.fetchall()[0]
        # working case
        posorient = pd.Series(index=['x', 'y', 'z',
                                     'alpha_0', 'alpha_1', 'alpha_2'])
        posorient.x = rows[5]
        posorient.y = rows[6]
        posorient.z = rows[1]
        posorient.alpha_0 = rows[3]
        posorient.alpha_1 = rows[2]
        posorient.alpha_2 = rows[4]
        image = pcode.scene(self.mydb, posorient=posorient)
        self.assertIsNotNone(image)
        self.assertFalse(sum(image.shape) == 0)
        # print("shape",image.shape)
        self.assertTrue(len(image.shape) == 4)
        self.assertTrue(image.shape[3] == 1)

        # incorrect case missing column
        posorient2 = pd.Series(index=['y', 'z',
                                      'alpha_0', 'alpha_1', 'alpha_2'])
        posorient2.y = posorient.y
        posorient2.z = posorient.z
        posorient2.alpha_0 = posorient.alpha_0
        posorient2.alpha_1 = posorient.alpha_1
        posorient2.alpha_2 = posorient.alpha_2
        with self.assertRaises(Exception):
            image = pcode.scene(self.mydb, posorient=posorient2)

        # incorrect case None
        posorient2 = pd.Series(index=['x', 'y', 'z',
                                      'alpha_0', 'alpha_1', 'alpha_2'])
        posorient2.x = None
        posorient2.y = posorient.y
        posorient2.z = posorient.z
        posorient2.alpha_0 = posorient.alpha_0
        posorient2.alpha_1 = posorient.alpha_1
        posorient2.alpha_2 = posorient.alpha_2
        with self.assertRaises(ValueError):
            image = pcode.scene(self.mydb, posorient=posorient2)

        # incorrect case nan
        posorient2 = pd.Series(index=['x', 'y', 'z',
                                      'alpha_0', 'alpha_1', 'alpha_2'])
        posorient2.x = np.nan
        posorient2.y = posorient.y
        posorient2.z = posorient.z
        posorient2.alpha_0 = posorient.alpha_0
        posorient2.alpha_1 = posorient.alpha_1
        posorient2.alpha_2 = posorient.alpha_2
        with self.assertRaises(ValueError):
            image = pcode.scene(self.mydb, posorient=posorient2)

        # incorrect case no pandas series but dict
        posorient2 = {}
        posorient2['x'] = posorient.x
        posorient2['y'] = posorient.y
        posorient2['z'] = posorient.z
        posorient2['alpha_0'] = posorient.alpha_0
        posorient2['alpha_1'] = posorient.alpha_1
        posorient2['alpha_2'] = posorient.alpha_2
        with self.assertRaises(TypeError):
            image = pcode.scene(self.mydb, posorient=posorient2)

        # not working case empty
        posorient2 = pd.Series(index=['x', 'y', 'z',
                                      'alpha_0', 'alpha_1', 'alpha_2'])

        with self.assertRaises(Exception):
            image = pcode.scene(self.mydb, posorient=posorient2)

    def test_skyline_scene(self):
        scene = pcode.scene(self.mydb, rowid=1)
        scene2 = scene.copy()
        scene2[3, 5, 2, 0] = np.nan
        scene3 = [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
        scene3 = [scene3, scene3, scene3]
        scene3 = np.array(scene3)
        scene4 = np.zeros((3, 4, 5, 0))

        # put useless stuff here
        with self.assertRaises(ValueError):
            pcode.skyline(scene2)
        with self.assertRaises(TypeError):
            pcode.skyline(scene3)
        with self.assertRaises(Exception):
            pcode.skyline(scene4)

        # should be working -> check if result(skyline) is correct
        for s in [scene]:
            skyline = pcode.skyline(s)
            self.assertFalse(skyline.shape[1] <= 0)
            self.assertTrue(skyline.shape[2] == 4)
            self.assertFalse(np.any(np.isnan(skyline)))
            # self.assertFalse(np.any(np.isNone(skyline)))
            self.assertTrue(is_numeric_array(skyline))
            self.assertTrue(skyline.shape[3] == 1)
            self.assertTrue(skyline.shape[0] > 0)
            self.assertTrue(skyline.shape[1] > 0)

    def test_id(self):
        for rowid in [0, -2]:
            with self.assertRaises(ValueError):
                # print("rowid",rowid)
                pcode.scene(self.mydb, rowid=rowid)
        with self.assertRaises(TypeError):
            pcode.scene(self.mydb, rowid='T')
        with self.assertRaises(ValueError):
            pcode.scene(self.mydb, rowid=None)
        with self.assertRaises(TypeError):
            pcode.scene(self.mydb, rowid=np.nan)
        with self.assertRaises(TypeError):
            pcode.scene(self.mydb, rowid=4.5)

        for rowid in [1, 2, 3, 4, 5]:
            image = pcode.scene(self.mydb, rowid=rowid)
            # image=np.array(image)
            self.assertIsNotNone(image)
            self.assertFalse(sum(image.shape) == 0)
            self.assertTrue(len(image.shape) == 4)
            self.assertFalse(np.any(np.isnan(image)))
            self.assertTrue(image.shape[3] == 1)
            self.assertTrue(image.shape[2] == 4)
            self.assertTrue(image.shape[0] > 0)
            self.assertTrue(image.shape[1] > 0)

    def test_distance_channel(self):
        scene = pcode.scene(self.mydb, rowid=1)
        # should not be working
        for d in ['g', None, np.nan, 8.4]:
            with self.assertRaises(TypeError):
                pcode.contrast_weighted_nearness(scene,
                                                 distance_channel=d)
        with self.assertRaises(ValueError):
            pcode.contrast_weighted_nearness(scene,
                                             distance_channel=-1)

        # should work
        d = 3
        weighted_scene = \
            pcode.contrast_weighted_nearness(scene,
                                             distance_channel=d)
        # print("last channel",d)
        self.assertTrue(is_numeric_array(weighted_scene))
        self.assertTrue(~np.any(np.isnan(weighted_scene)))
        self.assertEqual(weighted_scene.shape, scene.shape)

    def test_contr_weight_scene(self):
        scene = pcode.scene(self.mydb, rowid=1)

        # working cases
        contrast = pcode.contrast_weighted_nearness(scene)
        self.assertIsNotNone(contrast)
        self.assertFalse(sum(contrast.shape) == 0)
        self.assertTrue(len(contrast.shape) == 4)
        self.assertFalse(np.any(np.isnan(contrast)))
        self.assertTrue(contrast.shape[3] == 1)
        self.assertTrue(contrast.shape[2] == 4)
        self.assertTrue(contrast.shape[0] > 0)
        self.assertTrue(contrast.shape[1] > 0)

        # not working case
        scene2 = scene.copy()
        scene2[3, 2, 1, 0] = np.nan
        scene3 = [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
        scene3 = [scene3, scene3, scene3]
        scene3 = np.array(scene3)
        scene4 = np.zeros((3, 4, 5, 0))
        with self.assertRaises(ValueError):
            contrast = pcode.contrast_weighted_nearness(scene2)
        with self.assertRaises(Exception):
            contrast = pcode.contrast_weighted_nearness(scene3)
        with self.assertRaises(Exception):
            contrast = pcode.contrast_weighted_nearness(scene4)

    def test_contr_weight_contrast(self):
        scene = pcode.scene(self.mydb, rowid=1)
        for size in [9.4, 'g', None, np.nan]:
            with self.assertRaises(TypeError):
                contrast = pcode.contrast_weighted_nearness(
                    scene, contrast_size=size)
        for size in [8, 1, 0, -4]:
            with self.assertRaises(ValueError):
                contrast = \
                    pcode.contrast_weighted_nearness(
                        scene, contrast_size=size)

        # working cases
        for size in [2, 3, 4, 5]:
            contrast = pcode.contrast_weighted_nearness(scene,
                                                        contrast_size=size)
            self.assertIsNotNone(contrast)
            self.assertFalse(sum(contrast.shape) == 0)
            self.assertTrue(len(contrast.shape) == 4)
            self.assertFalse(np.any(np.isnan(contrast)))
            self.assertEqual(contrast.shape[3], 1)
            self.assertEqual(contrast.shape[2], scene.shape[2])
            self.assertEqual(contrast.shape[0], scene.shape[0])
            self.assertEqual(contrast.shape[1], scene.shape[1])

    def test_pcv(self):
        # working case
        rowid = 1
        my_scene = pcode.scene(self.mydb, rowid=rowid)
        directions = self.mydb.viewing_directions
        my_pcv = pcode.pcv(my_scene, directions)
        self.assertIsNotNone(my_pcv)
        self.assertFalse(sum(my_pcv.shape) == 0)
        self.assertTrue(len(my_pcv.shape) == 4)
        self.assertFalse(np.any(np.isnan(my_pcv)))
        self.assertTrue(my_pcv.shape[3] == 3)
        self.assertTrue(my_pcv.shape[2] == 4)
        self.assertTrue(my_pcv.shape[0] > 0)
        self.assertTrue(my_pcv.shape[1] > 0)

        # not working cases doesnt match with shape of place code
        testdirection = np.zeros((2, 4, 2))
        with self.assertRaises(Exception):
            my_pcv = pcode.pcv(my_scene, testdirection)

        # not working cases wrong last dimension
        testdirection = np.zeros((180, 360, 1))
        with self.assertRaises(Exception):
            my_pcv = pcode.pcv(my_scene, testdirection)

        # not working cases too many dimensions
        testdirection = np.zeros((180, 360, 2, 4))
        with self.assertRaises(Exception):
            my_pcv = pcode.pcv(my_scene, testdirection)

        # not working cases empty
        testdirection = np.zeros(())
        with self.assertRaises(Exception):
            my_pcv = pcode.pcv(my_scene, testdirection)

        # not working cases nans
        testdirection = np.zeros((180, 360, 2, 4))
        testdirection[2, 3, 0] = np.nan
        with self.assertRaises(ValueError):
            my_pcv = pcode.pcv(my_scene, testdirection)

    def test_apcv(self):
        # working case
        rowid = 1
        my_scene = pcode.scene(self.mydb, rowid=rowid)
        # print("scene shape",my_scene.shape)
        directions = self.mydb.viewing_directions
        # print("directions",directions.shape)
        my_pcv = pcode.apcv(my_scene, directions)

        self.assertIsNotNone(my_pcv)
        self.assertFalse(sum(my_pcv.shape) == 0)
        self.assertTrue(len(my_pcv.shape) == 4)
        self.assertFalse(np.any(np.isnan(my_pcv)))
        self.assertTrue(my_pcv.shape[3] == 3)
        self.assertTrue(my_pcv.shape[2] == 4)
        self.assertTrue(my_pcv.shape[0] == 1)
        self.assertTrue(my_pcv.shape[1] == 1)

        # not working cases doesnt match with shape of place code
        testdirection = np.zeros((2, 4, 2))
        with self.assertRaises(Exception):
            my_pcv = pcode.apcv(my_scene, testdirection)

        # not working cases wrong last dimension
        testdirection = np.zeros((180, 360, 1))
        with self.assertRaises(Exception):
            my_pcv = pcode.apcv(my_scene, testdirection)

        # not working cases too many dimensions
        testdirection = np.zeros((180, 360, 2, 4))
        with self.assertRaises(Exception):
            my_pcv = pcode.apcv(my_scene, testdirection)

        # not working cases empty
        testdirection = np.zeros(())
        with self.assertRaises(Exception):
            my_pcv = pcode.apcv(my_scene, testdirection)

        # not working cases nans
        testdirection = np.zeros((180, 360, 2, 4))
        testdirection[2, 3, 0] = np.nan
        with self.assertRaises(ValueError):
            my_pcv = pcode.apcv(my_scene, testdirection)

    def test_size(self):
        # not working cases:
        scene = pcode.scene(self.mydb, rowid=1)
        for size in [8, 1, 0, -4]:
            with self.assertRaises(ValueError):
                contrast = pcode.michelson_contrast(
                    scene, size=size)
        for size in [9.4, 'g', None, np.nan]:
            with self.assertRaises(TypeError):
                contrast = pcode.michelson_contrast(
                    scene, size=size)

        # working cases
        for size in [2, 3, 4, 5]:
            contrast = pcode.michelson_contrast(scene, size=size)
            self.assertIsNotNone(contrast)
            self.assertFalse(sum(contrast.shape) == 0)
            self.assertTrue(len(contrast.shape) == 4)
            self.assertFalse(np.any(np.isnan(contrast)))
            self.assertTrue(contrast.shape[3] == 1)
            self.assertTrue(contrast.shape[2] == 4)
            self.assertTrue(contrast.shape[0] > 0)
            self.assertTrue(contrast.shape[1] > 0)

    def test_michelsoncontrast_scene(self):

        scene = pcode.scene(self.mydb, rowid=1)

        # working cases
        contrast = pcode.michelson_contrast(scene)
        self.assertIsNotNone(contrast)
        self.assertFalse(sum(contrast.shape) == 0)
        self.assertTrue(len(contrast.shape) == 4)
        self.assertFalse(np.any(np.isnan(contrast)))
        self.assertTrue(contrast.shape[3] == 1)
        self.assertTrue(contrast.shape[2] == 4)
        self.assertTrue(contrast.shape[0] > 0)
        self.assertTrue(contrast.shape[1] > 0)

        # not working case
        scene2 = scene.copy()
        scene2[3, 2, 1, 0] = np.nan
        scene3 = [[1, 2, 3], [1, 2, 3], [1, 2, 3]]
        scene3 = [scene3, scene3, scene3]
        scene3 = np.array(scene3)
        scene4 = np.zeros((3, 4, 5, 0))
        for s in [scene2, scene3, scene4]:
            with self.assertRaises(Exception) as cm:
                contrast = pcode.michelson_contrast(s,)
            print("wanted exception occured", cm.exception)


if __name__ == '__main__':
    unittest.main()
