from .constants import __ibpc_indeces__
from .constants import __spherical_indeces__
from .constants import __cartesian_indeces__
from .constants import __obpc_indeces__
from .constants import __eye_indeces__
import numpy as np


def check_scene(scene):
    if is_ibpc(scene):
        #  print("normal")
        if not is_numeric_array(scene):
            raise TypeError('scene is of non numeric type')
        if not ~np.any(np.isnan(scene)):
            raise ValueError('scene contains nans')
        if not len(scene.shape) == 4:
            raise Exception('scene has wrong shape, must have 4 dimensions')
        if not (scene.shape[1] > 0):
            raise Exception('scenes first dimension is empty')
        if not (scene.shape[0] > 0):
            raise Exception('scenes second dimension is empty')
        if not (scene.shape[2] == 4):
            raise Exception('3rd dimension of scene must be four')
        if not (scene.shape[3] == 1):
            raise Exception('4rd dimension of scene must be one')
        # assert ~(np.any(np.isNone(scene)))
        return True

    elif is_obpc(scene):
        if not is_numeric_array(scene):
            raise TypeError('scene is of non numeric type')
        if not ~np.any(np.isnan(scene)):
            raise ValueError('scene contains nans')
        if not len(scene.shape) == 3:
            raise Exception('scene has wrong shape, must have 4 dimensions')
        if not ~(scene.shape[1] <= 0):
            raise Exception('scenes first dimension is empty')
        if not ~(scene.shape[0] <= 0):
            raise Exception('scenes second dimension is empty')
        if not (scene.shape[2] == 4):
            raise Exception('3rd dimension of scene must be four')
        if not (scene.shape[3] == 1):
            raise Exception('4rd dimension of scene must be one')
        # assert ~(np.any(np.isNone(scene)))
        return True


def check_viewing_direction(viewing_direction):
    if not is_numeric_array(viewing_direction):
        raise TypeError('viewing direction is of non numeric type')
    if not ~np.any(np.isnan(viewing_direction)):
        raise ValueError('viewing direction contains nans')
    if len(viewing_direction.shape) < 3:
        raise Exception('viewing direction must have at least 3 dimensions')
    if not (viewing_direction.shape[1] > 0):
        raise Exception('viewing direction has empty second dimension')
    if not (viewing_direction.shape[0] > 0):
        raise Exception('viewing direction has empty first dimension')
    if not (viewing_direction.shape[-1] == 2):
        raise Exception(' last dimension of viewing direction must equal 2')
    return True


def is_numeric_array(array):
    """Checks if the dtype of the array is numeric.

    Booleans, unsigned integer, signed integer, floats and complex are
    considered numeric.

    :param array : `numpy.ndarray`-like The array to check.
    :returns: True if it is a recognized numerical and False \
    if object or  string.
    :rtype:bool
    """
    numerical_dtype_kinds = {'b',  # boolean
                             'u',  # unsigned integer
                             'i',  # signed integer
                             'f',  # floats
                             'c'}  # complex
    try:
        return array.dtype.kind in numerical_dtype_kinds
    except AttributeError:
        # in case it's not a numpy array it will probably have no dtype.
        return np.asarray(array).dtype.kind in numerical_dtype_kinds


def is_ibpc(place_code):
    """Test if a place code is image based

    :param place_code: a place-code
    :returns: True if image based place-code
    :rtype: bool

    """
    toreturn = isinstance(place_code, np.ndarray)
    toreturn = toreturn and (len(place_code.shape) ==
                             len(__ibpc_indeces__))
    return toreturn


def is_obpc(place_code):
    """Test if a place code is ommatidia based

    :param place_code: a place-code
    :returns: True if ommatidia based place-code
    :rtype: bool

    """
    toreturn = isinstance(place_code, np.ndarray)
    toreturn = toreturn and (len(place_code.shape) ==
                             len(__obpc_indeces__))
    return toreturn


def ibs_to_obs(scene, eye_map):
    """Convert an image based scene to an ommatidium based scene.

    :param scene: The scene to be converted
    :param eye_map: The eye_map to use
    :returns: (obs_scene,ommatidia_map)
    :rtype: (np.ndarray,np.ndarray)
    """
    assert is_ibpc(scene),\
        'scene should be an ibs scene'
    assert isinstance(eye_map, np.ndarray), 'eye_map should be a numpy array'
    assert len(eye_map.shape) == len(__eye_indeces__),\
        'eye_map should have {} dimensions to be an ibs scene'.format(
            len(__eye_indeces__))
    for index_name in ['elevation', 'azimuth']:
        index = __ibpc_indeces__[index_name]
        assert eye_map.shape[index] == scene.shape[index],\
            'eye_map and scene should have the same number of {}'.format(
                index_name)
    obs_size = (scene.shape[__ibpc_indeces__['elevation']] *
                scene.shape[__ibpc_indeces__['azimuth']],
                scene.shape[__ibpc_indeces__['channel']],
                scene.shape[__ibpc_indeces__['component']])
    obs_scene = scene.reshape(obs_size)
    omm_size = (eye_map.shape[__ibpc_indeces__['elevation']] *
                eye_map.shape[__ibpc_indeces__['azimuth']],
                eye_map.shape[__ibpc_indeces__['component']])
    ommatidia_map = eye_map.reshape(omm_size)
    return (obs_scene, ommatidia_map)


def cartesian_to_spherical(x, y, z):
    radius = np.sqrt(x**2 + y**2 + z**2)
    elevation = np.arctan2(z, np.sqrt(x**2 + y**2))
    azimuth = np.arctan2(y, x)
    spherical = np.zeros_like(x)
    spherical = np.tile(spherical[..., np.newaxis], (3,))
    spherical[..., __spherical_indeces__['elevation']] = elevation
    spherical[..., __spherical_indeces__['azimuth']] = azimuth
    spherical[..., __spherical_indeces__['radius']] = radius
    return spherical


def spherical_to_cartesian(elevation, azimuth, radius=1):
    cartesian = np.zeros_like(elevation)
    cartesian = np.tile(cartesian[..., np.newaxis], (3,))
    cartesian[..., __cartesian_indeces__['x']] = np.cos(
        elevation) * np.cos(azimuth)
    cartesian[..., __cartesian_indeces__['y']] = np.cos(
        elevation) * np.sin(azimuth)
    cartesian[..., __cartesian_indeces__['z']] = np.sin(elevation)
    cartesian = radius * cartesian
    return cartesian
