"""
The beesampling class

.. tothinkof: conditional bpy import to build doc from comments
"""
import bpy
import os
import numpy as np
import pandas as pd
import warnings
from navipy.rendering.cyber_bee import Cyberbee
from navipy.database import DataBaseSave


class BeeSampling(Cyberbee):
    """
    BeeSampling is a class deriving from Cyberbee.
    The BeeSampling can be used to generate a database of
images taken on a rectangular regular grid. For the database,
the BeeSampling rely on DataBase
    It worth noting that the generated database can take a large
harddrive space, as each image is composed of 4 channels of 180x360 pixels.
    """

    def __init__(self):
        """Initialise the BeeSampling"""
        Cyberbee.__init__(self)
        self.blenddirname = os.path.dirname(bpy.data.filepath)
        self.blendfilename = os.path.basename(bpy.data.filepath)
        self.__grid_posorients = None
        self.__grid_size = None
        self.world_dim = np.inf

    def create_sampling_grid(self, x, y, z, alpha1, alpha2, alpha3):
        """Create a cubic grid from all the sampling points

        :param x: the positions along the x-axis
        :param y: the positions along the y-axis
        :param z: the positions along the z-axis
        :param alpha1: the first euler angles
        :param alpha2: the 2nd euler angles
        :param alpha3: the 3rd euler angles
        """
        if not (isinstance(x, int) or isinstance(x, float)):
            raise TypeError('x must be integer')
        if not (isinstance(y, int) or isinstance(y, float)):
            raise TypeError('y must be integer')
        if not (isinstance(z, int) or isinstance(z, float)):
            raise TypeError('z must be integer')
        if not (isinstance(alpha1, float) or isinstance(alpha1, int)):
            raise TypeError('alpha1 must be float')
        if not (isinstance(alpha2, float) or isinstance(alpha2, int)):
            raise TypeError('alpha2 must be float')
        if not (isinstance(alpha3, float) or isinstance(alpha3, int)):
            raise TypeError('alpha3 must be float')
        [mx, my, mz, ma1, ma2, ma3] = np.meshgrid(x,
                                                  y,
                                                  z,
                                                  alpha1,
                                                  alpha2,
                                                  alpha3)
        self.grid_size = mx.shape
        mx = mx.flatten()
        my = my.flatten()
        mz = mz.flatten()
        ma1 = ma1.flatten()
        ma2 = ma2.flatten()
        ma3 = ma3.flatten()
        self.__grid_posorients = pd.DataFrame(index=range(mx.shape[0]),
                                              columns=['x', 'y', 'z',
                                                       'alpha_0',
                                                       'alpha_1',
                                                       'alpha_2'])
        self.__grid_posorients.loc[:, 'x'] = mx
        self.__grid_posorients.loc[:, 'y'] = my
        self.__grid_posorients.loc[:, 'z'] = mz
        self.__grid_posorients.loc[:, 'alpha_0'] = ma1
        self.__grid_posorients.loc[:, 'alpha_1'] = ma2
        self.__grid_posorients.loc[:, 'alpha_2'] = ma3
        self.__grid_posorients.index.name = 'grid_index'
        self.__grid_posorients.name = 'grid_position_orientation'

    @property
    def grid_posorients(self):
        """return a copy of the posorientation matrix

        :returns: position-orientations of the grid
        :rtype: pandas array

        .. todo: use @property
                     def grid_posorients(self)

        .. todo: create @property.setter
                        def grid_posorients(self,posorients)
                        (need a type check, and col check, and copy of df)
        """
        return self.__grid_posorients.copy()

    def set_gridindeces2nan(self, indeces):
        """Set certain grid point to nan, so they will be ignore in the rendering

        :param indeces: a list of indeces to be set to nan

        .. todo: use @property.setter
                     def blacklist_indeces(self,indeces)
        """
        if not isinstance(indeces, list):
            raise TypeError('indeces must be a list')
        if any(np.isnan(indeces)):
            raise ValueError('indeces must not contain nans')
        self.__grid_posorients.loc[indeces, :] = np.nan

    def render(self, database_filename):
        if not isinstance(database_filename, str):
            raise TypeError('filename must be a string')
        database_folder = os.path.dirname(database_filename)
        if not os.path.exists(database_folder):
            os.makedirs(database_folder)
        dataloger = DataBaseSave(database_filename,
                                 channels=['R', 'G', 'B', 'D'],
                                 arr_dtype=np.uint8)
        for frame_i, posorient in self.__grid_posorients.iloc[::-1].iterrows():
            print(frame_i)
            if np.any(np.isnan(posorient)):
                # warnings.warn('frame_i: {} posorient nans'.format(frame_i))
                continue
            rowid = dataloger.posid(posorient)
            if dataloger.check_data_validity(rowid):
                warnings.warn(
                    'frame_i: {} data is valid rowid {}'.format(frame_i,
                                                                rowid))
                continue
            # The position-orientatios is valid (none nan)
            # and the cmaxminrange has not already been assigned
            # so the image need to be rendered
            self.update(posorient)
            distance = self.distance
            distance[distance > self.world_dim] = self.world_dim
            image = self.image
            image[:, :, 3] = distance
            dataloger.write_image(posorient, image)
        print('rendering completed')


if __name__ == "__main__":
    import tempfile
    bee_samp = BeeSampling()
    # Create mesh
    world_dim = 15.0
    x = np.linspace(-7.5, 7.5, 5)
    y = np.linspace(-7.5, 7.5, 5)
    z = np.arange(1, 8, 2)
    alpha_1 = np.array([0]) + np.pi / 2
    alpha_2 = np.array([0])
    alpha_3 = np.array([0])
    bee_samp.create_sampling_grid(
        x, y, z, alpha1=alpha_1, alpha2=alpha_2, alpha3=alpha_3)
    bee_samp.world_dim = world_dim
    grid_pos = bee_samp.grid_posorients
    condition = (grid_pos.x**2 + grid_pos.y**2) < ((bee_samp.world_dim / 2)**2)
    bee_samp.set_gridindeces2nan(condition[condition == 0].index)
    bee_samp.cycle_samples(samples=5)
    with tempfile.TemporaryDirectory() as folder:
        bee_samp.render(folder + '/database.db')
